package com.sillycycle.bagleyd.rubik;

/*
 * @(#)Rubik.java
 *
 * Copyright 1994 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>Rubik</code> class extends the
 * <code>RubikCanvas</code> for handling the responses to the user
 * for the <code>RubikFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

/*
The Simple Solutions to Cubic Puzzles by James G. Nourse, Bantam Books,
New York, 1981. pp 8-15.

Mastering the Magic Pyramid by Tom Werneck, Evans Brothers Limited, London,
1981. pp 109-111.

Beyond Rubik's Cube: spheres, pyramids, dodecahedrons and God knows what
else by Douglas R. Hofstadter, Scientific American, July 1982, pp 16-31.

Magic Cubes 1996 Catalog of Dr. Christoph Bandelow.
*/

public class Rubik extends Canvas implements MouseWheelListener,
		MouseListener, MouseMotionListener, KeyListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_ITERATIONS = 100;

	static final int AMBIGUOUS = -3;
	static final int IGNORE = -1;
	static final int TOP = 0;
	static final int RIGHT = 1;
	static final int BOTTOM = 2;
	static final int LEFT = 3;
	static final int STRT = 4;
	static final int CW = 5;
	static final int HALF = 6;
	static final int CCW = 7;
	static final int MAX_FACES = 6;

	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final String DATAFILE = "rubik.dat";
	static final String SYMBOL = ":";

	Color       foreground, background;
	Color       borderColor;
	Color       currentForeground;
	Color[]     faceColor = new Color[MAX_FACES];
	int[]       directionOfPosition = null;
	int         currentDirection = -1;
	int         currentFace = -1, nextFace = -1;
	int         currentPosition = -1, nextPosition = -1;
	int         currentRotate = -1, currentH = -1;
	boolean     currentReverse = false;
	boolean     currentControl = false, nextControl = false;
	int         rowPosition;
	boolean     started, cheat;
	boolean     mono, reverse, orientChoice, practiceChoice;
	int         delay;
	int         sizeX, sizeY, sizeZ, dim;
	int         lastX, lastY;

	JFrame      frame = null;
	boolean     mouseDown = false;
	Random      generator;
	RubikStack  undo = null, redo = null;
	RubikSolve  solve = null;
	Rubik2D	    rubik2D = null;
	Rubik3D	    rubik3D = null;
	String      stringSave = null;
	String[]    token = null;
	String      faceNames = null;

	static final int[][] facesToDirection/*[MAX_FACES][MAX_FACES]*/ = {
		{-1, 3, 2, 1, -3, 0},
		{0, -1, 1, -3, 2, 3},
		{0, 3, -1, 1, 2, -3},
		{0, -3, 3, -1, 2, 1},
		{-3, 3, 0, 1, -1, 2},
		{2, 3, -3, 1, 0, -1}
	};
	RubikLoc[][] slideNextRow = new RubikLoc[MAX_FACES][RubikInterface.MAX_ORIENT];
	static final int[][] rowToRotate/*[MAX_FACES][RubikInterface.MAX_ORIENT]*/ = {
		{3, 2, 1, 5},
		{2, 4, 5, 0},
		{3, 4, 1, 0},
		{5, 4, 2, 0},
		{3, 5, 1, 2},
		{3, 0, 1, 4}
	};
	RubikLoc[][] cubeLoc = null;
	static RubikLoc[][] startLoc = null;
	RubikLoc[] faceLoc = null;
	RubikLoc[][] rowLoc = null;

	public Rubik(JFrame frame, int dim,
			int sizeX, int sizeY, int sizeZ,
			boolean orient, boolean practice,
			boolean mono, boolean reverse, String names,
			Color c0, Color c1, Color c2,
			Color c3, Color c4, Color c5,
			Color borderc, Color fg, Color bg, int delay) {
		this.frame = frame;
		this.dim = dim;
		this.sizeX = sizeX;
		this.sizeY = sizeY;
		this.sizeZ = sizeZ;
		this.orientChoice = orient;
		this.practiceChoice = practice;
		this.mono = mono;
		this.reverse = reverse;
		this.faceNames = names;
		this.faceColor[0] = c0;
		this.faceColor[1] = c1;
		this.faceColor[2] = c2;
		this.faceColor[3] = c3;
		this.faceColor[4] = c4;
		this.faceColor[5] = c5;
		this.currentForeground = this.borderColor = borderc;
		this.foreground = fg; this.background = bg;
		this.delay = delay;
		rubik2D = new Rubik2D(frame, this);
		rubik3D = new Rubik3D(frame, this);
		undo = ((RubikFrame) frame).undo;
		redo = ((RubikFrame) frame).redo;
		if (dim == 2)
			rubik2D.firstPaint = true;
		else if (dim == 3)
			rubik3D.firstPaint = true;
		this.addMouseListener(this);
		this.addMouseMotionListener(this);
		this.addMouseWheelListener(this);
		this.addKeyListener(this);
		setCursor(new Cursor(Cursor.HAND_CURSOR));
	}

	@Override
	protected String paramString() {
		return super.paramString() +
		",sizeX=" + sizeX +
		",sizeY=" + sizeY +
		",sizeZ=" + sizeZ +
		",orient=" + orientChoice +
		",practice=" + practiceChoice +
		",mono=" + mono +
		",reverse=" + reverse +
		",names=" + faceNames +
		",color0=" + faceColor[0] +
		",color1=" + faceColor[1] +
		",color2=" + faceColor[2] +
		",color3=" + faceColor[3] +
		",color4=" + faceColor[4] +
		",color5=" + faceColor[5] +
		",color6=" + faceColor[6] +
		",color7=" + faceColor[7] +
		",borderColor=" + borderColor +
		",foreground=" + foreground +
		",background=" + background +
		",delay=" + delay;
	}

	void checkPieces() {
		if (sizeX < RubikInterface.MIN_CUBES) {
			System.out.println(sizeX +
				" number of cubes on a edge out of bounds, must be at least " +
				RubikInterface.MIN_CUBES +
				", defaulting to " +
				RubikInterface.DEFAULTCUBES);
			sizeX = RubikInterface.DEFAULTCUBES;
		}
		if (sizeY < RubikInterface.MIN_CUBES) {
			System.out.println(sizeY +
				" number of cubes on a edge out of bounds, must be at least " +
				RubikInterface.MIN_CUBES +
				", defaulting to " +
				RubikInterface.DEFAULTCUBES);
			sizeY = RubikInterface.DEFAULTCUBES;
		}
		if (sizeZ < RubikInterface.MIN_CUBES) {
			System.out.println(sizeZ +
				" number of cubes on a edge out of bounds, must be at least " +
				RubikInterface.MIN_CUBES +
				", defaulting to " +
				RubikInterface.DEFAULTCUBES);
			sizeZ = RubikInterface.DEFAULTCUBES;
		}
		if (delay < 0) {
			System.out.println(delay +
			" delay cannot be negative, taking absolute value");
			delay = -delay;
		}
	}

	int sizeRow(int face) {
		switch (face) {
		case 0: /* TOP */
		case 4: /* BOTTOM */
			return sizeX;
		case 1: /* LEFT */
		case 3: /* RIGHT */
			return sizeZ;
		case 2: /* FRONT */
		case 5: /* BACK */
			return sizeX;
		default:
			System.out.println("sizeRow: face " + face);
			return 0;
		}
	}

	int sizeColumn(int face) {
		switch (face) {
		case 0: /* TOP */
		case 4: /* BOTTOM */
			return sizeZ;
		case 1: /* LEFT */
		case 3: /* RIGHT */
			return sizeY;
		case 2: /* FRONT */
		case 5: /* BACK */
			return sizeY;
		default:
			System.out.println("sizeColumn: face " + face);
			return 0;
		}
	}

	int sizeFace(int face) {
		return (sizeRow(face) * sizeColumn(face));
	}

	boolean checkFaceSquare(int face) {
		int sizeOfRow = sizeRow(face);
		int sizeOfColumn = sizeColumn(face);

		return (sizeOfRow == sizeOfColumn);
		/* Cubes can be made square with a 4x2 face where 90 degree turns
		 * should be permitted but that is kind of complicated for me.
		 * This can be done in 2 ways where the side of the cubies are
		 * the same size and one where one side (the side with half the
		 * number of cubies) is twice the size of the other.  The first is
		 * complicated because faces of cubies can go under other faces.
		 * The second way is similar to "banded cubes" where scotch tape
		 * restricts the moves of some cubes.  Here you have to keep track
		 * of the restrictions and show banded cubies graphically as one
		 * cube.
		 */
	}

	int maxArea() {
		return sizeX * sizeY * sizeZ / java.lang.Math.min(sizeX,
			java.lang.Math.min(sizeY, sizeZ));
	}

	int maxLen() {
		return java.lang.Math.max(sizeX,
			java.lang.Math.max(sizeY, sizeZ));
	}

	boolean checkStarted() {
		return started;
	}

	boolean checkSolved(boolean rotation) {
		for (int face = 0; face < MAX_FACES; face++) {
			int tempFace = cubeLoc[face][0].face;
			int tempRotation = cubeLoc[face][0].rotation;

			for (int position = 1; position < sizeFace(face);
					position++) {
				if (tempFace != /* face */
						cubeLoc[face][position].face ||
						(rotation && tempRotation != /* STRT * MAX_ORIENT */
						cubeLoc[face][position].rotation))
					return false;
			}
		}
		return true;
	}

	void drawSquare(int face, int position, int offsetPosition) {
		if (dim == 2)
			rubik2D.drawSquare(face, position, offsetPosition);
		else if (dim == 3)
			rubik3D.drawSquare(face, position, offsetPosition);
	}

	void drawFace(int face) {
		for (int position = 0; position < sizeFace(face); position++)
			drawSquare(face, position, 0);
	}

	void drawAllPieces() {
		for (int face = 0; face < MAX_FACES; face++)
			drawFace(face);
		solve.init();
	}

	void moveNoPieces() {
		((RubikFrame) frame).callback(
			RubikInterface.ACTION_ILLEGAL, dim);
	}

	void rotateFace(int face, int direction) {
		int i, j;
		int sizeOfRow = sizeRow(face);
		int sizeOfColumn = sizeColumn(face);
		int sizeOnPlane = sizeOfRow * sizeOfColumn;
		faceLoc = new RubikLoc[sizeOnPlane];

		/* Read Face */
		for (int position = 0; position < sizeOnPlane; position++) {
			faceLoc[position] = new RubikLoc(cubeLoc[face][position].face,
				cubeLoc[face][position].rotation);
		}
		/* Write Face */
		for (int position = 0; position < sizeOnPlane; position++) {
			i = position % sizeOfRow;
			j = position / sizeOfRow;
			switch (direction) {
			case CW:
				cubeLoc[face][position].face =
					faceLoc[(sizeOfRow - i - 1) * sizeOfRow + j].face;
				cubeLoc[face][position].rotation =
					faceLoc[(sizeOfRow - i - 1) * sizeOfRow + j].rotation;
				break;
			case CCW:
				cubeLoc[face][position].face =
					faceLoc[i * sizeOfRow + sizeOfColumn - j - 1].face;
				cubeLoc[face][position].rotation =
					faceLoc[i * sizeOfRow + sizeOfColumn - j - 1].rotation;
				break;
			default:
				/* (direction == HALF) */
				cubeLoc[face][position].face =
					faceLoc[sizeOfRow - i - 1 + (sizeOfColumn - j - 1) * sizeOfRow].face;
				cubeLoc[face][position].rotation =
					faceLoc[sizeOfRow - i - 1 + (sizeOfColumn - j - 1) * sizeOfRow].rotation;
				break;
			}
			cubeLoc[face][position].rotation =
				(cubeLoc[face][position].rotation +
				 direction - RubikInterface.MAX_ORIENT) % RubikInterface.MAX_ORIENT;
			drawSquare(face, position, 0);
		}
	}

	void slideRC(int face, int direction, int h, int sizeOnOppAxis) {
		currentFace = slideNextRow[face][direction].face;
		currentRotate = slideNextRow[face][direction].rotation;
		currentDirection = (currentRotate + direction) % RubikInterface.MAX_ORIENT;
		switch (currentRotate) {
		case TOP:
			currentH = h;
			currentReverse = false;
			break;
		case RIGHT:
			if (currentDirection == TOP || currentDirection == BOTTOM) {
				currentH = sizeOnOppAxis - 1 - h;
				currentReverse = false;
			} else { /* currentDirection == RIGHT ||
					currentDirection == LEFT */
				currentH = h;
				currentReverse = true;
			}
			break;
		case BOTTOM:
			currentH = sizeOnOppAxis - 1 - h;
			currentReverse = true;
			break;
		case LEFT:
			if (currentDirection == TOP || currentDirection == BOTTOM) {
				currentH = h;
				currentReverse = true;
			} else {	/* currentDirection == RIGHT ||
					   currentDirection == LEFT */
				currentH = sizeOnOppAxis - 1 - h;
				currentReverse = false;
			}
			break;
		default:
			System.out.println("slideRC: rotate " + currentRotate);
		}
	}

	void readRC(int face, int dir, int h, int orient, int size) {
		int sizeOfRow;

		sizeOfRow = sizeRow(face);
		if (dir == TOP || dir == BOTTOM)
			for (int g = 0; g < size; g++) {
				rowLoc[orient][g].face =
					cubeLoc[face][g * sizeOfRow + h].face;
				rowLoc[orient][g].rotation =
					cubeLoc[face][g * sizeOfRow + h].rotation;
			}
		else /* dir == RIGHT || dir == LEFT */
			for (int g = 0; g < size; g++) {
				rowLoc[orient][g].face =
					cubeLoc[face][h * sizeOfRow + g].face;
				rowLoc[orient][g].rotation =
					cubeLoc[face][h * sizeOfRow + g].rotation;
			}
	}

	void rotateRC(int rotate, int orient, int size) {
		for (int g = 0; g < size; g++)
			rowLoc[orient][g].rotation =
				(rowLoc[orient][g].rotation + rotate) % RubikInterface.MAX_ORIENT;
	}

	void reverseRC(int orient, int size) {
		RubikLoc temp;

		for (int g = 0; g < size / 2; g++) {
			temp = rowLoc[orient][size - 1 - g];
			rowLoc[orient][size - 1 - g] = rowLoc[orient][g];
			rowLoc[orient][g] = temp;
		}
	}

	void writeRC(int face, int dir, int h, int orient, int size) {
		int position, sizeOfRow;

		sizeOfRow = sizeRow(face);
		if (dir == TOP || dir == BOTTOM) {
			for (int g = 0; g < size; g++) {
				position = g * sizeOfRow + h;
				cubeLoc[face][position].face = rowLoc[orient][g].face;
				cubeLoc[face][position].rotation = rowLoc[orient][g].rotation;
				drawSquare(face, position, 0);
			}
		} else { /* dir == RIGHT || dir == LEFT */
			for (int g = 0; g < size; g++) {
				position = h * sizeOfRow + g;
				cubeLoc[face][position].face = rowLoc[orient][g].face;
				cubeLoc[face][position].rotation = rowLoc[orient][g].rotation;
				drawSquare(face, position, 0);
			}
		}
	}

	void movePieces(int face, int position, int direction) {
		int dir = direction, oldFace = face;
		int h, k;
		int i, j, sizeOnAxis, sizeOnOppAxis;
		int sizeOfRow = sizeRow(oldFace);
		int sizeOfColumn = sizeColumn(oldFace);
		int degreeTurn = (checkFaceSquare(
			rowToRotate[oldFace][dir])) ? 90 : 180;

		currentReverse = false;
		currentH = 0;
		i = position % sizeOfRow;
		j = position / sizeOfRow;
		h = ((dir == TOP || dir == BOTTOM) ? i : j);
		if (dir == TOP || dir == BOTTOM) {
			sizeOnAxis = sizeOfColumn;
			sizeOnOppAxis = sizeOfRow;
		} else {
			sizeOnAxis = sizeOfRow;
			sizeOnOppAxis = sizeOfColumn;
		}
		/* rotate sides CW or CCW or HALF */

		if (h == sizeOnOppAxis - 1) {
			currentDirection = ((dir == TOP || dir == BOTTOM) ?
				TOP : RIGHT);
			if (degreeTurn == 180)
				rotateFace(rowToRotate[oldFace][currentDirection], HALF);
			else if (dir == TOP || dir == RIGHT)
				rotateFace(rowToRotate[oldFace][currentDirection], CW);
			else		/* dir == BOTTOM || dir == LEFT */
				rotateFace(rowToRotate[oldFace][currentDirection], CCW);
		}
		if (h == 0) {
			currentDirection = ((dir == TOP || dir == BOTTOM) ?
				BOTTOM : LEFT);
			if (degreeTurn == 180)
				rotateFace(rowToRotate[oldFace][currentDirection], HALF);
			else if (dir == TOP || dir == RIGHT)
				rotateFace(rowToRotate[oldFace][currentDirection], CCW);
			else		/* dir == BOTTOM  || dir == LEFT */
				rotateFace(rowToRotate[oldFace][currentDirection], CW);
		}
		/* Slide rows */
		readRC(oldFace, dir, h, 0, sizeOnAxis);
		if (degreeTurn == 180) {
			int sizeOnDepthAxis;

			slideRC(oldFace, dir, h, sizeOnOppAxis); /* sets currentFace,
				currentDirection, currentH, currentRotate, currentReverse */
			sizeOnDepthAxis = sizeFace(currentFace) / sizeOnOppAxis;
			readRC(currentFace, currentDirection, currentH, 1, sizeOnDepthAxis);
			rotateRC(currentRotate, 0, sizeOnAxis);
			if (currentReverse)
				reverseRC(0, sizeOnAxis);
			oldFace = currentFace;
			dir = currentDirection;
			h = currentH;
			for (k = 2; k <= RubikInterface.MAX_ORIENT + 1; k++) {
				slideRC(oldFace, dir, h, sizeOnOppAxis); /* sets currentFace,
					currentDirection, currentH, currentRotate, currentReverse */
				if (k != RubikInterface.MAX_ORIENT && k != RubikInterface.MAX_ORIENT + 1)
					readRC(currentFace, currentDirection, currentH, k,
						 (((k & 1) == 1) ? sizeOnDepthAxis : sizeOnAxis));
				rotateRC(currentRotate, k - 2,
					 (((k & 1) == 1) ? sizeOnDepthAxis : sizeOnAxis));
				if (k != RubikInterface.MAX_ORIENT + 1)
					rotateRC(currentRotate, k - 1,
						 (((k & 1) == 1) ? sizeOnAxis : sizeOnDepthAxis));
				if (currentReverse) {
					reverseRC(k - 2,
						 (((k & 1) == 1) ? sizeOnDepthAxis : sizeOnAxis));
					if (k != RubikInterface.MAX_ORIENT + 1)
						reverseRC(k - 1,
							  (((k & 1) == 1) ? sizeOnAxis : sizeOnDepthAxis));
				}
				writeRC(currentFace, currentDirection, currentH, k - 2,
					(((k & 1) == 1) ? sizeOnDepthAxis : sizeOnAxis));
				oldFace = currentFace;
				dir = currentDirection;
				h = currentH;
			}
		} else {
			for (k = 1; k <= RubikInterface.MAX_ORIENT; k++) {
				slideRC(oldFace, dir, h, sizeOnOppAxis); /* sets currentFace,
					currentDirection, currentH, currentRotate, currentReverse */
				if (k != RubikInterface.MAX_ORIENT)
					readRC(currentFace, currentDirection, currentH, k, sizeOnAxis);
				rotateRC(currentRotate, k - 1, sizeOnAxis);
				if (currentReverse)
					reverseRC(k - 1, sizeOnAxis);
				writeRC(currentFace, currentDirection, currentH, k - 1, sizeOnAxis);
				oldFace = currentFace;
				dir = currentDirection;
				h = currentH;
			}
		}
	}

	void moveControlCb(int face, int direction) {
		int position, sizeOnOppAxis, factor;
		int sizeOfRow = sizeRow(face);
		int sizeOfColumn = sizeColumn(face);

		if (direction == TOP || direction == BOTTOM) {
			sizeOnOppAxis = sizeOfRow;
			factor = 1;
		} else {
			sizeOnOppAxis = sizeOfColumn;
			factor = sizeOfRow;
		}
		for (int k = 0; k < sizeOnOppAxis; k++) {
			position = k * factor;
			movePieces(face, position, direction);
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_CONTROL, dim,
				face, position, direction);
		}
	}

	void movePuzzle(int face, int position, int direction,
			boolean control) {
		if (control) {
			moveControlCb(face, direction);
		} else {
			movePieces(face, position, direction);
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_MOVED, dim,
				face, position, direction);
		}
		undo.setMove(direction, (control) ? 1 : 0,
			face, position);
		redo.flushMoves();
		if (((RubikFrame) frame).getToggleSound())
			((RubikFrame) frame).playMoveAudio();
	}

	void movePuzzleDelay(int face, int position, int direction,
			boolean control, int fast) {
		movePuzzle(face, position, direction, control);
		try {
			Thread.sleep(64 * delay / fast);
		} catch (InterruptedException e) {
			// ignore
		}
	}

	void selectPieces(/*int x, int y*/) {
		if (!nextControl && !practiceChoice && checkSolved(orientChoice))
			moveNoPieces();
		else if (nextFace >= 0) {
			int count;

			if (nextFace == currentFace) {
				// next sets currentDirection
				count = checkMoveDir(currentFace,
					currentPosition, nextPosition);
				/* count == 2 too annoying */
				if (count == 2)
					count = -1;
			} else {
				currentDirection =
					facesToDirection[currentFace][nextFace];
				count = (currentDirection == AMBIGUOUS) ?
					2 : 1;
				/* can look closer to minimize this, but then
				must also be able to handle double moves */
			}
			switch (count) {
			case 1:
				movePuzzle(currentFace, currentPosition,
					currentDirection, nextControl);
				started = true;
				if (!nextControl && checkSolved(orientChoice)) {
					((RubikFrame) frame).callback(
						RubikInterface.ACTION_SOLVED,
						dim);
				}
				break;
			case 2:
				((RubikFrame) frame).callback(
					RubikInterface.ACTION_AMBIGUOUS, dim);
				break;
			case 0:
				moveNoPieces();
				break;
			default:
				break;
			}
		}
	}

	int checkMoveDir(int face, int position1, int position2) {
		int	 count = 0;
		int	 column1, column2, row1, row2, sizeOfRow;

		sizeOfRow = sizeRow(face);
		column1 = position1 % sizeOfRow;
		column2 = position2 % sizeOfRow;
		row1 = position1 / sizeOfRow;
		row2 = position2 / sizeOfRow;
		if (column1 == column2 && row1 != row2) {
			currentDirection = (row2 > row1) ? BOTTOM : TOP;
			count = 1;
		} else if (row1 == row2 && column1 != column2) {
			currentDirection = (column2 > column1) ? RIGHT : LEFT;
			count = 1;
		} else if (row1 == row2) // column1 == column2
			count = 2;
		return count;
	}

	boolean narrowSelection() {
		// sets currentFace, currentPosition, currentDirection
		if (dim == 2)
			rubik2D.narrowSelection();
		else if (dim == 3)
			rubik3D.narrowSelection();
		return true;
	}

	boolean positionPieces(int x, int y, int direction) {
		// sets currentFace, currentPosition
		if (dim == 2) {
			currentFace = rubik2D.positionToPiece(x, y);
		} else if (dim == 3) {
			currentFace = rubik3D.positionToPiece(x, y);
		}
		if (currentFace < 0)
			return false;
		currentDirection = direction;
		narrowSelection(); // sets currentFace, currentPosition, currentDirection
		return (currentFace >= 0);
	}

	void movePuzzleInput(int x, int y, int direction,
			boolean control) {
		if (!practiceChoice && !control && checkSolved(orientChoice)) {
			moveNoPieces();
			return;
		}
		if (!positionPieces(x, y, direction)) // sets currentFace, currentPosition, currentDirection
			return;
		movePuzzle(currentFace, currentPosition, currentDirection,
			control);
		if (!control && checkSolved(orientChoice)) {
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_SOLVED, dim);
		}
	}

	void resetPieces() {
		int sizeSize = maxArea();
		int size = maxLen();

		cubeLoc = new RubikLoc[MAX_FACES][sizeSize];
		startLoc = new RubikLoc[MAX_FACES][sizeSize];
		rowLoc = new RubikLoc[RubikInterface.MAX_ORIENT][size];
		for (int face = 0; face < MAX_FACES; face++) {
			int sizeOfFace = sizeFace(face);
			for (int position = 0; position < sizeOfFace; position++) {
				startLoc[face][position] = new RubikLoc(face,
					STRT - RubikInterface.MAX_ORIENT);
				cubeLoc[face][position] = new RubikLoc(face,
					STRT - RubikInterface.MAX_ORIENT);
			}
		}
		for (int orient = 0; orient < RubikInterface.MAX_ORIENT; orient++)
			for (int position = 0; position < size; position++)
				rowLoc[orient][position] = new RubikLoc(0, 0);
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentFace = -1;
		currentPosition = -1;
		currentDirection = -1;
		started = false;
		cheat = false;
		rubik2D.randomPaints = 0;
		rubik3D.randomPaints = 0;
	}

	void practicePieces() {
		((RubikFrame) frame).callback(
			RubikInterface.ACTION_PRACTICE, dim);
	}

	void randomizePieces() {
		if (practiceChoice)
			practicePieces();
		((RubikFrame) frame).callback(
			RubikInterface.ACTION_RESET, dim);
		int big = sizeX * sizeY * sizeZ * 3;
		if (dim == 2)
			rubik2D.randomPaints = java.lang.Math.min(big,
				MAX_ITERATIONS) + generator.nextInt(2);
		else if (dim == 3)
			rubik3D.randomPaints = java.lang.Math.min(big,
				MAX_ITERATIONS) + generator.nextInt(2);
		repaint();
	}

	void randomizingPieces() {
		int position, randomDirection;
		int face = generator.nextInt(MAX_FACES);

		randomDirection = generator.nextInt(RubikInterface.MAX_ORIENT);
		position = generator.nextInt(sizeFace(face));
		movePuzzle(face, position, randomDirection, false);
		if (dim == 2)
			rubik2D.randomPaints--;
		else if (dim == 3)
			rubik3D.randomPaints--;
		try {
			Thread.sleep(delay);
		} catch (InterruptedException e) {
			// ignore
		}
		if (dim == 2 && rubik2D.randomPaints > 0) {
			repaint();
		} else if (dim == 3 && rubik3D.randomPaints > 0) {
			repaint();
		} else {
			undo.flushMoves();
			redo.flushMoves();
			getStartPosition();
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_RANDOMIZE, dim);
			started = false;
			cheat = false;
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_CHEAT, dim);
			currentFace = -1;
			if (checkSolved(orientChoice))
				((RubikFrame) frame).callback(
					RubikInterface.ACTION_SOLVED, dim);
		}
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
	}

	public boolean getCheat() {
		return cheat;
	}

	public void setCheat(boolean cheater) {
		cheat = cheater;
	}

	public void setStart(boolean start) {
		started = start;
	}

	boolean setRubikSizeX(int newXValue) {
		if (sizeX != newXValue &&
				newXValue >= RubikInterface.MIN_CUBES) {
			sizeX = newXValue;
			resetPieces();
			if (dim == 2)
				rubik2D.resizePaint = true;
			else if (dim == 3)
				rubik3D.resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setRubikSizeY(int newYValue) {
		if (sizeY != newYValue &&
				newYValue >= RubikInterface.MIN_CUBES) {
			sizeY = newYValue;
			resetPieces();
			if (dim == 2)
				rubik2D.resizePaint = true;
			else if (dim == 3)
				rubik3D.resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setRubikSizeZ(int newZValue) {
		if (sizeZ != newZValue &&
				newZValue >= RubikInterface.MIN_CUBES) {
			sizeZ = newZValue;
			resetPieces();
			if (dim == 2)
				rubik2D.resizePaint = true;
			else if (dim == 3)
				rubik3D.resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setRubikOrient(boolean newValue) {
		if (orientChoice != newValue) {
			orientChoice = newValue;
			resetPieces();
			if (dim == 2)
				rubik2D.resizePaint = true;
			else if (dim == 3)
				rubik3D.resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setRubikPractice(boolean newValue) {
		if (practiceChoice != newValue) {
			practiceChoice = newValue;
			resetPieces();
			if (dim == 2)
				rubik2D.resizePaint = true;
			else if (dim == 3)
				rubik3D.resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public void speedUpPuzzle() {
		delay -= 10;
		if (delay < 0)
			delay = 0;
		((RubikFrame) frame).callback(RubikInterface.ACTION_SPEED, dim);
	}

	public void slowDownPuzzle() {
		delay += 10;
		((RubikFrame) frame).callback(RubikInterface.ACTION_SPEED, dim);
	}

	void showMessage(String msg) {
		((RubikFrame) frame).showMessage(msg);
	}

	public void initializePuzzle() {
		checkPieces();
		solve = new RubikSolve(frame, this);
		resetPieces();
		generator = new Random(System.nanoTime());
		slideNextRow[0][0] = new RubikLoc(5, TOP);
		slideNextRow[0][1] = new RubikLoc(3, RIGHT);
		slideNextRow[0][2] = new RubikLoc(2, TOP);
		slideNextRow[0][3] = new RubikLoc(1, LEFT);
		slideNextRow[1][0] = new RubikLoc(0, RIGHT);
		slideNextRow[1][1] = new RubikLoc(2, TOP);
		slideNextRow[1][2] = new RubikLoc(4, LEFT);
		slideNextRow[1][3] = new RubikLoc(5, BOTTOM);
		slideNextRow[2][0] = new RubikLoc(0, TOP);
		slideNextRow[2][1] = new RubikLoc(3, TOP);
		slideNextRow[2][2] = new RubikLoc(4, TOP);
		slideNextRow[2][3] = new RubikLoc(1, TOP);
		slideNextRow[3][0] = new RubikLoc(0, LEFT);
		slideNextRow[3][1] = new RubikLoc(5, BOTTOM);
		slideNextRow[3][2] = new RubikLoc(4, RIGHT);
		slideNextRow[3][3] = new RubikLoc(2, TOP);
		slideNextRow[4][0] = new RubikLoc(2, TOP);
		slideNextRow[4][1] = new RubikLoc(3, LEFT);
		slideNextRow[4][2] = new RubikLoc(5, TOP);
		slideNextRow[4][3] = new RubikLoc(1, RIGHT);
		slideNextRow[5][0] = new RubikLoc(4, TOP);
		slideNextRow[5][1] = new RubikLoc(3, BOTTOM);
		slideNextRow[5][2] = new RubikLoc(0, TOP);
		slideNextRow[5][3] = new RubikLoc(1, BOTTOM);
		if (dim == 2) {
			rubik2D.resizePuzzle();
		} else if (dim == 3) {
			rubik3D.resizePuzzle();
		}
	}

	void exposePuzzle() {
		if (dim == 2)
			rubik2D.exposePuzzle();
		else if (dim == 3)
			rubik3D.exposePuzzle();
	}

	void selectPuzzle(boolean control, int x, int y) {
		if (dim == 2)
			rubik2D.selectPuzzle(control, x, y);
		else if (dim == 3)
			rubik3D.selectPuzzle(control, x, y);
	}

	void releasePuzzle(boolean control, int x, int y) {
		if (dim == 2)
			rubik2D.releasePuzzle(control, x, y);
		else if (dim == 3)
			rubik3D.releasePuzzle(control, x, y);
	}

	void randomizePuzzle() {
		if (!mouseDown)
			randomizePieces();
	}

	void randomizePuzzleWithQuery() {
		if (!started)
			randomizePuzzle();
	}

	public void getPuzzle(boolean incMoves) {
		this.stringSave = ((RubikFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			if (incMoves)
				showMessage("Empty string to read");
			return;
		}
		getPuzzleString("buffer", incMoves);
	}

	public void getPuzzleFile(boolean incMoves) {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();

		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.length() == 0) {
				if (incMoves)
					showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE, incMoves);
		} catch (Exception e) {
			//e.printStackTrace();
			if (incMoves)
				showMessage("Cannot read " + DATAFILE);
		}
	}

	 public void getPuzzleString(String source, boolean incMoves) {
		try {
			int i, newSizeX, newSizeY, newSizeZ, moves;
			boolean orient, practice;
			String symbol = SYMBOL;
			String newline = NL;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				if (incMoves)
					showMessage("nothing in buffer");
				return;
			}
			newSizeX = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (newSizeX >= RubikInterface.MIN_CUBES) {
				for (i = this.sizeX; i < newSizeX; i++) {
					((RubikFrame) frame).callback(
						RubikInterface.ACTION_INCX, dim);
				}
				for (i = this.sizeX; i > newSizeX; i--) {
					((RubikFrame) frame).callback(
						RubikInterface.ACTION_DECX, dim);
				}
			} else {
				if (incMoves)
					showMessage(source +
						" corrupted: sizeX " + newSizeX +
						" should be at least " +
						RubikInterface.MIN_CUBES);
				return;
			}
			lineData = token[++line];
			newSizeY = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (newSizeY >= RubikInterface.MIN_CUBES) {
				for (i = this.sizeY; i < newSizeY; i++) {
					((RubikFrame) frame).callback(
						RubikInterface.ACTION_INCY, dim);
				}
				for (i = this.sizeY; i > newSizeY; i--) {
					((RubikFrame) frame).callback(
						RubikInterface.ACTION_DECY, dim);
				}
			} else {
				if (incMoves)
					showMessage(source +
						" corrupted: sizeY " + newSizeY +
						" should be at least " +
						RubikInterface.MIN_CUBES);
				return;
			}
			lineData = token[++line];
			newSizeZ = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (newSizeZ >= RubikInterface.MIN_CUBES) {
				for (i = this.sizeZ; i < newSizeZ; i++) {
					((RubikFrame) frame).callback(
						RubikInterface.ACTION_INCZ, dim);
				}
				for (i = this.sizeZ; i > newSizeZ; i--) {
					((RubikFrame) frame).callback(
						RubikInterface.ACTION_DECZ, dim);
				}
			} else {
				if (incMoves)
					showMessage(source +
						" corrupted: sizeZ " + newSizeZ +
						" should be at least " +
						RubikInterface.MIN_CUBES);
				return;
			}
			lineData = token[++line];
			orient = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.orientChoice != orient)
				((RubikFrame) frame).callback(
					RubikInterface.ACTION_ORIENTIZE, dim);
			lineData = token[++line];
			practice = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.practiceChoice != practice)
				((RubikFrame) frame).callback(
					RubikInterface.ACTION_PRACTICE, dim);
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line, incMoves);
			if (line < 0)
				return;
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_RESTORE, dim);
			setStartPosition();
			if (dim == 2) {
				rubik2D.allPiecesPaint = true;
			} else if (dim == 3) {
				rubik3D.allPiecesPaint = true;
			}
			repaint();
			if (!scanMoves(++line, moves, incMoves))
				return;
			if (incMoves)
				System.out.println(source + ": sizeX " + newSizeX +
					", sizeY " + newSizeY + ", sizeZ " + newSizeZ +
					", orient " + orient +
					", practice " + practice + ", moves " + moves);
			cheat = true;
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_CHEAT, dim);
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			if (incMoves)
				showMessage("Corrupt input when reading");
		} catch (Exception e) {
			//e.printStackTrace();
			if (incMoves)
				showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("sizeX").append(symbol).append(space)
				.append(sizeX).append(newline);
		sb.append("sizeY").append(symbol).append(space)
				.append(sizeY).append(newline);
		sb.append("sizeZ").append(symbol).append(space)
				.append(sizeZ).append(newline);
		sb.append("orient").append(symbol).append(space)
				.append((orientChoice) ? 1 : 0).append(newline);
		sb.append("practice").append(symbol).append(space)
				.append((practiceChoice) ? 1 : 0).append(newline);
		sb.append("moves").append(symbol).append(space)
				.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((RubikFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	void undoPuzzle() {
		if (mouseDown || rubik2D.stackPaint || rubik3D.stackPaint ||
			rubik2D.randomPaints != 0 || rubik3D.randomPaints != 0 ||
			!undo.madeMoves())
			return;
		undo.getMove();
		currentDirection = undo.readMoveDirection();
		currentControl = (undo.readMoveControl() != 0);
		currentFace = undo.readMoveFace();
		currentPosition = undo.readMovePosition();
		redo.setMove(currentDirection, (currentControl) ? 1 : 0,
			currentFace, currentPosition);
		// move in opposite direction
		currentDirection =
			(currentDirection < RubikInterface.MAX_ORIENT) ?
			(currentDirection + RubikInterface.MAX_ORIENT / 2) %
			RubikInterface.MAX_ORIENT :
			3 * RubikInterface.MAX_ORIENT - currentDirection;
		if (currentControl) {
			if (dim == 2)
				rubik2D.stackPaint = true;
			else if (dim == 3)
				rubik3D.stackPaint = true;
			repaint();
		} else /* if (checkmovePiecesDir(currentDirection)) */ {
			if (dim == 2)
				rubik2D.stackPaint = true;
			else if (dim == 3)
				rubik3D.stackPaint = true;
			repaint();
			started = true;
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_UNDO, dim,
				currentFace, currentPosition,
				currentDirection);
		} /* else currentDirection = currentFace = -1; */
	}

	void redoPuzzle() {
		if (mouseDown || rubik2D.stackPaint || rubik3D.stackPaint ||
			rubik2D.randomPaints != 0 || rubik3D.randomPaints != 0 ||
			!redo.madeMoves())
			return;
		redo.getMove();
		currentDirection = redo.readMoveDirection();
		currentControl = (redo.readMoveControl() != 0);
		currentFace = redo.readMoveFace();
		currentPosition = redo.readMovePosition();
		undo.setMove(currentDirection, (currentControl) ? 1 : 0,
			currentFace, currentPosition);
		if (currentControl) {
			if (dim == 2)
				rubik2D.stackPaint = true;
			else if (dim == 3) 
				rubik3D.stackPaint = true;
			repaint();
		} else /* if (checkmovePiecesDir(currentDirection)) */ {
			if (dim == 2)
				rubik2D.stackPaint = true;
			else if (dim == 3)
				rubik3D.stackPaint = true;
			repaint();
			started = true;
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_REDO, dim,
				currentFace, currentPosition,
				currentDirection);
		} /* else currentDirection = currentFace = -1; */
	}

	public void clearPuzzle() {
		if (mouseDown || rubik2D.randomPaints != 0 ||
				rubik3D.randomPaints != 0)
			return;
		resetPieces();
		drawAllPieces();
		((RubikFrame) frame).callback(RubikInterface.ACTION_CLEAR, dim);
		if (dim == 2)
			rubik2D.allPiecesPaint = true;
		else if (dim == 3)
			rubik3D.allPiecesPaint = true;
		repaint();
	}

	void solvePuzzle() {
		if (mouseDown || rubik2D.stackPaint || rubik3D.stackPaint ||
			rubik2D.randomPaints != 0 || rubik3D.randomPaints != 0)
			return;
		if ((sizeX == 2 && sizeY == 2 && sizeZ == 2) ||
			(sizeX == 3 && sizeY == 3 && sizeZ == 3) ||
			(sizeX == 4 && sizeY == 4 && sizeZ == 4)) {
			cheat = true;
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_CHEAT, dim);
			solve.setFlag(true);
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {
				System.err.println("Interrupted");
			}
			//solve.solvePieces();
		} else {
			showMessage("Auto-solver: sorry, only implemented for size" +
				NL + "nxnxn where n < 5.");
		}
	}

	void incrementXPuzzle(boolean shift) {
		if (mouseDown || rubik2D.randomPaints != 0 ||
				rubik3D.randomPaints != 0)
			return;
		if (shift) {
			if (sizeX <= RubikInterface.MIN_CUBES)
				return;
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_DECX, dim);
		} else {
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_INCX, dim);
		}
	}

	void incrementYPuzzle(boolean shift) {
		if (mouseDown || rubik2D.randomPaints != 0 ||
				rubik3D.randomPaints != 0)
			return;
		if (shift) {
			if (sizeY <= RubikInterface.MIN_CUBES)
				return;
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_DECY, dim);
		} else {
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_INCY, dim);
		}
	}

	void incrementZPuzzle(boolean shift) {
		if (mouseDown || rubik2D.randomPaints != 0 ||
				rubik3D.randomPaints != 0)
			return;
		if (shift) {
			if (sizeZ <= RubikInterface.MIN_CUBES)
				return;
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_DECZ, dim);
		} else {
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_INCZ, dim);
		}
	}

	void incrementPuzzle() {
		if (mouseDown || rubik2D.randomPaints != 0 ||
				rubik3D.randomPaints != 0)
			return;
		((RubikFrame) frame).callback(
			RubikInterface.ACTION_INCX, dim);
		((RubikFrame) frame).callback(
			RubikInterface.ACTION_INCY, dim);
		((RubikFrame) frame).callback(
			RubikInterface.ACTION_INCZ, dim);
	}

	void decrementPuzzle() {
		if (mouseDown || rubik2D.randomPaints != 0 ||
				rubik3D.randomPaints != 0)
			return;
		if (sizeX > RubikInterface.MIN_CUBES)
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_DECX, dim);
		if (sizeY > RubikInterface.MIN_CUBES)
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_DECY, dim);
		if (sizeZ > RubikInterface.MIN_CUBES)
			((RubikFrame) frame).callback(
				RubikInterface.ACTION_DECZ, dim);
	}

	void orientizePuzzle() {
		if (mouseDown || rubik2D.randomPaints != 0 ||
				rubik3D.randomPaints != 0)
			return;
		((RubikFrame) frame).callback(
			RubikInterface.ACTION_ORIENTIZE, dim);
	}

	void practicePuzzle() {
		if (mouseDown || rubik2D.randomPaints != 0 ||
				rubik3D.randomPaints != 0)
			return;
		practicePieces();
	}

	void practicePuzzleWithQuery() {
		if (!started)
			practicePuzzle();
	}

	void printPositions() {
		for (int face = 0; face < MAX_FACES; face++) {
			int sizeOfColumn = sizeColumn(face);
			int sizeOfFace = sizeFace(face);

			for (int position = 0; position < sizeOfFace; position++)
			{
				int pos = startLoc[face][position].face;
				int rot = startLoc[face][position].rotation;

				if (pos < 100)
					System.out.print(" ");
				if (pos < 10 && pos >= 0)
					System.out.print(" ");
				System.out.print(pos + "," + rot);
				if (position % sizeOfColumn == sizeOfColumn - 1) {
					System.out.println("");
				} else {
					System.out.print(" ");
				}
			}
			System.out.println("");
		}
	}

	void printCube() {
		int sizeOfRow, sizeOfColumn;

		for (int face = 0; face < MAX_FACES; face++) {
			sizeOfRow = sizeRow(face);
			sizeOfColumn = sizeColumn(face);
			for (int g = 0; g < sizeOfRow; g++) {
				for (int h = 0; h < sizeOfColumn; h++)
					System.out.print(
						cubeLoc[face][h + g * sizeOfColumn].face + " " +
						cubeLoc[face][h + g * sizeOfColumn].rotation +
						"  ");
			}
			System.out.println("");
		}
		System.out.println("");
	}

	void printFace(int face) {
		int sizeOfRow = sizeRow(face);
		int sizeOfColumn = sizeColumn(face);

		for (int g = 0; g < sizeOfRow; g++) {
			for (int h = 0; h < sizeOfColumn; h++)
				System.out.print(
					faceLoc[h + g * sizeOfColumn].face + " " +
					faceLoc[h + g * sizeOfColumn].rotation +
					"  ");
			System.out.println("");
		}
		System.out.println("");
	}


	void printRow(int orient, int size) {
		System.out.println("Row " + orient + ":");
		for (int g = 0; g < size; g++)
			System.out.print(
					rowLoc[orient][g].face + " " +
					rowLoc[orient][g].rotation +
					"  ");
		System.out.println("");
	}

	void getStartPosition() {
		for (int face = 0; face < MAX_FACES; face++)
			for (int position = 0; position < sizeFace(face); position++)
			{
				startLoc[face][position].face =
					cubeLoc[face][position].face;
				startLoc[face][position].rotation =
					cubeLoc[face][position].rotation;
			}
	}

	void setStartPosition() {
		for (int face = 0; face < MAX_FACES; face++)
			for (int position = 0; position < sizeFace(face); position++)
			{
				cubeLoc[face][position].face =
					startLoc[face][position].face;
				cubeLoc[face][position].rotation =
					startLoc[face][position].rotation;
			}
		// drawAll
	}

	int scanStartPosition(int location, boolean incMoves) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";
			int group = (orientChoice) ? 2 : 1;

			++line; //read startingPosition:
			for (int face = 0; face < MAX_FACES; face++) {
				int pos = 0;
				int faceOfSize = sizeFace(face);
				++line;
				while (token.length > line && token[line].equals(""))
					++line;
				lineData = token[line];
				String[] st = lineData.split(space);
				int offset = ("".equals(st[0])) ? 1 : 0;
				for (int position = 0; position < faceOfSize; position++)
				{
					int num = Integer.parseInt(st[group * pos + offset].trim());
					cubeLoc[face][position].face =
					startLoc[face][position].face = num;
					if (orientChoice) {
						num = Integer.parseInt(st[group * pos + offset + 1].trim());
						cubeLoc[face][position].rotation =
						startLoc[face][position].rotation = num;
					}
					int width;
					if (face == 1 || face == 3) {
						width = this.sizeZ;
					} else {
						width = this.sizeX;
					}
					if ((position + 1) % width == 0 &&
							(position + 1) < faceOfSize) {
						line++;
						if (line >= token.length) {
							if (incMoves)
								System.out.println("Short data in StartPosition");
							return -1;
						}
						lineData = token[line];
						st = lineData.split(space);
						pos = 0;
					} else
						pos++;
				}
			}
			//blank line read
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			if (incMoves)
				showMessage("Corrupt input in StartPosition");
			return -1;
		}
		return line;
	}

	boolean scanMoves(int location, int moves, boolean incMoves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String symbol = SYMBOL;
			String space = "\\s+";

			++line; //read moves direction
			paint(g);
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData =  token[line];
				String[] st = lineData.substring(
						lineData.indexOf(symbol) + 2).split(space);
				currentDirection = Integer.parseInt(st[0].trim());
				currentControl = (Integer.parseInt(st[1].trim()) != 0);
				currentFace = Integer.parseInt(st[2].trim());
				currentPosition = Integer.parseInt(st[3].trim());
				{
					if (dim == 2 && !currentControl) {
						rubik2D.scanPaint = true;
						paint(g);
					} else if (dim == 3) {
						rubik3D.scanPaint = true;
						undo.setMove(currentDirection,
							(currentControl) ? 1 : 0,
							currentFace, currentPosition);
						redo.flushMoves();
						paint(g);
					}
					if (!currentControl)
						((RubikFrame) frame).callback(
						  RubikInterface.ACTION_MOVED,
						  dim);
				} /* else {
					currentDirection = -1;
					System.out.println(l +
						": move in direction " +
						currentDirection +
						", cannot be made.");
					return false;
				} */
			}
			if (l < moves) {
				if (incMoves)
					showMessage("Corrupt input in scanMoves, only " + l + " moves");
				return false;
			}
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			if (incMoves)
				showMessage("Corrupt input in scanMoves");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int face = 0; face < MAX_FACES; face++) {
			for (int position = 0; position < sizeFace(face); position++)
			{
				int temp = startLoc[face][position].face;
				if (temp < 100)
					sb.append(space);
				if (temp < 10 && temp >= 0)
					sb.append(space);
				sb.append(space).append(temp);
				if (orientChoice) {
					temp = startLoc[face][position].rotation;
					if (temp < 10 && temp >= 0)
						sb.append(space);
					sb.append(space).append(temp);
				}
				int width;
				if (face == 1 || face == 3) {
					width = this.sizeZ;
				} else {
					width = this.sizeX;
				}
				if ((position + 1) % width == 0) {
					sb.append(newline);
				}
			}
			sb.append(newline);
		}
	}

	void printMoves(StringBuilder sb) {
		int direction, control;
		int face, position;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space)
			.append("dir").append(space)
			.append("control").append(space)
			.append("face").append(space)
			.append("position").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			direction = undo.readMoveDirection();
			control = undo.readMoveControl();
			face = undo.readMoveFace();
			position = undo.readMovePosition();
			redo.setMove(direction, control, face, position);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			direction = redo.readMoveDirection();
			control = redo.readMoveControl();
			face = redo.readMoveFace();
			position = redo.readMovePosition();
			sb.append(i + 1).append(symbol).append(space)
				.append(direction).append(space)
				.append(control).append(space)
				.append(face).append(space)
				.append(position).append(newline);
			undo.setMove(direction, control, face, position);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (dim == 2)
			rubik2D.paint(g);
		else if (dim == 3)
			rubik3D.paint(g);
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	public boolean processKey(char key) {
		switch (key) {
		case '>':
		case '.':
			speedUpPuzzle();
			return true;
		case '<':
		case ',':
			slowDownPuzzle();
			return true;
		case '@':
			((RubikFrame) frame).toggleSound();
			return true;
		case '?':
			((RubikFrame) frame).descriptionHelp();
			return true;
		case '!':
			((RubikFrame) frame).featuresHelp();
			return true;
		case '^':
			((RubikFrame) frame).referencesHelp();
			return true;
		case 'a':
		case 'A':
			((RubikFrame) frame).aboutHelp();
			return true;
		case '*':
			((RubikFrame) frame).showBuffer();
			return true;
		default:
		}
		if (solve != null && solve.getFlag()) {
			solve.setFlag(false);
			return false;
		}
		switch (key) {
		case 'g':
			((RubikFrame) frame).getPuzzle();
			break;
		case 'w':
			writePuzzle();
			break;
		case 'G':
			((RubikFrame) frame).getPuzzleFile();
			break;
		case 'W':
			writePuzzleFile();
			break;
		case 'u':
		case 'U':
			undoPuzzle();
			break;
		case 'r':
		case 'R':
			redoPuzzle();
			break;
		case 'c':
		case 'C':
			clearPuzzle();
			break;
		case 'z':
		case 'Z':
			randomizePuzzle();
			break;
		case 'p':
		case 'P':
			practicePuzzle();
			break;
		case 's':
		case 'S':
			solvePuzzle();
			break;
		case 'o':
		case 'O':
			orientizePuzzle();
			break;
		case 'i':
		case 'I':
			incrementPuzzle();
			break;
		case 'd':
		case 'D':
			decrementPuzzle();
			break;
		case 'q':
		case 'Q':
		case 'x':
		case 'X':
		case '\03': /* Control-C */
			((RubikFrame) frame).quit();
			break;
		default:
			return false;
		}
		return true;
	}

	@Override
	public void mousePressed(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if (rubik2D.scanPaint || rubik3D.scanPaint ||
				rubik2D.stackPaint || rubik3D.stackPaint || mouseDown)
			return;
		if (solve != null && solve.getFlag()) {
			solve.setFlag(false);
			return;
		}
		if (((modifiers & InputEvent.META_DOWN_MASK) == 0) &&
				((modifiers & InputEvent.ALT_DOWN_MASK) == 0)) {
			if (rubik2D.randomPaints == 0 &&
					rubik3D.randomPaints == 0) {
				selectPuzzle((modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
					event.getX(), event.getY());
				mouseDown = true;
			}
		} else {
			if ((event.getClickCount() & 1) == 1)
				randomizePuzzleWithQuery();
			else
				randomizePuzzle();
		}
	}

	@Override
	public void mouseClicked(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	@Override
	public void mouseReleased(MouseEvent event) {
		int modifiers = event.getModifiersEx();

		if (mouseDown) { // Do not care which mouse button
			releasePuzzle((modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				event.getX(), event.getY());
			mouseDown = false;
		}
	}

	@Override
	public void mouseEntered(MouseEvent event) {
		currentForeground = foreground;
		if (dim == 2) {
			/* rubik2D.requestFocus(); */
			rubik2D.framePaint = true;
		}
		else if (dim == 3) {
			/* rubik3D.requestFocus(); */
			rubik3D.framePaint = true;
		}
		mouseDown = false;
		repaint();
	}

	@Override
	public void mouseExited(MouseEvent event) {
		if (mouseDown && currentFace >= 0) {
			if (dim == 2)
				rubik2D.releasePaint = true;
			else if (dim == 3)
				rubik3D.releasePaint = true;
		}
		currentForeground = borderColor;
		if (dim == 2)
			rubik2D.framePaint = true;
		else if (dim == 3)
			rubik3D.framePaint = true;
		mouseDown = false;
		repaint();
	}

	@Override
	public void mouseDragged(MouseEvent event) {
		/* Method required but mousePressed does work */
	}

	@Override
	public void mouseMoved(MouseEvent event) {
		lastX = event.getX();
		lastY = event.getY();
	}

	@Override
	public void mouseWheelMoved(MouseWheelEvent event) {
		int notches = event.getWheelRotation();
		int modifiers = event.getModifiersEx();

		if (notches < 0) {
			movePuzzleInput(lastX, lastY, TOP,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
		} else {
			movePuzzleInput(lastX, lastY, BOTTOM,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0);
		}
	}

	@Override
	public void keyPressed(KeyEvent event) {
		if (event.getKeyCode() == KeyEvent.VK_ESCAPE)
			((RubikFrame) frame).shuffleDown();
		else if (!processKey(event.getKeyChar())) {
			if (dim == 2)
				rubik2D.keyPressed(event);
			else if (dim == 3)
				rubik3D.keyPressed(event);
		}
	}

	@Override
	public void keyReleased(KeyEvent event) {
		/* Method required but keyPressed does work */
	}

	@Override
	public void keyTyped(KeyEvent event) {
		/* Method required but keyPressed does work */
	}
}

/**
 * The <code>RubikLoc</code> class maintains the locations of
 * the pieces.
 */

class RubikLoc {
	int		 face;
	int		 rotation;

	public RubikLoc(int face, int rotation) {
		this.face = face;
		this.rotation = rotation;
	}
}
